@@ -23,6 +23,7 @@ import os
 
 from novaclient import api_versions
 from novaclient.v2 import servers
+from openstack import exceptions as sdk_exceptions
 from osc_lib.cli import parseractions
 from osc_lib.command import command
 from osc_lib import exceptions
@@ -251,13 +252,16 @@ class AddFloatingIP(network_common.NetworkAndComputeCommand):
         parser.add_argument(
             "ip_address",
             metavar="<ip-address>",
-            help=_("Floating IP address to assign to server (IP only)"),
+            help=_("Floating IP address to assign to the first available "
+                   "server port (IP only)"),
         )
         parser.add_argument(
             "--fixed-ip-address",
             metavar="<ip-address>",
             help=_(
-                "Fixed IP address to associate with this floating IP address"
+                "Fixed IP address to associate with this floating IP address. "
+                "The first server port containing the fixed IP address will "
+                "be used"
             ),
         )
         return parser
@@ -274,12 +278,45 @@ class AddFloatingIP(network_common.NetworkAndComputeCommand):
             compute_client.servers,
             parsed_args.server,
         )
-        port = list(client.ports(device_id=server.id))[0]
-        attrs['port_id'] = port.id
+        ports = list(client.ports(device_id=server.id))
+        # If the fixed IP address was specified, we need to find the
+        # corresponding port.
         if parsed_args.fixed_ip_address:
-            attrs['fixed_ip_address'] = parsed_args.fixed_ip_address
-
-        client.update_ip(obj, **attrs)
+            fip_address = parsed_args.fixed_ip_address
+            attrs['fixed_ip_address'] = fip_address
+            for port in ports:
+                for ip in port.fixed_ips:
+                    if ip['ip_address'] == fip_address:
+                        attrs['port_id'] = port.id
+                        break
+                else:
+                    continue
+                break
+            if 'port_id' not in attrs:
+                msg = _('No port found for fixed IP address %s')
+                raise exceptions.CommandError(msg % fip_address)
+            client.update_ip(obj, **attrs)
+        else:
+            # It's possible that one or more ports are not connected to a
+            # router and thus could fail association with a floating IP.
+            # Try each port until one succeeds. If none succeed, re-raise the
+            # last exception.
+            error = None
+            for port in ports:
+                attrs['port_id'] = port.id
+                try:
+                    client.update_ip(obj, **attrs)
+                except sdk_exceptions.NotFoundException as exp:
+                    # 404 ExternalGatewayForFloatingIPNotFound from neutron
+                    LOG.info('Skipped port %s because it is not attached to '
+                             'an external gateway', port.id)
+                    error = exp
+                    continue
+                else:
+                    error = None
+                    break
+            if error:
+                raise error
 
     def take_action_compute(self, client, parsed_args):
         client.api.floating_ip_add(
@@ -809,9 +846,14 @@ class CreateServer(command.ShowOne):
                     raise exceptions.CommandError(msg)
                 nics = nics[0]
         else:
-            # Default to empty list if nothing was specified, let nova side to
-            # decide the default behavior.
-            nics = []
+            # Compute API version >= 2.37 requires a value, so default to
+            # 'auto' to maintain legacy behavior if a nic wasn't specified.
+            if compute_client.api_version >= api_versions.APIVersion('2.37'):
+                nics = 'auto'
+            else:
+                # Default to empty list if nothing was specified, let nova
+                # side to decide the default behavior.
+                nics = []
 
         # Check security group exist and convert ID to name
         security_group_names = []
@@ -1045,11 +1087,22 @@ class ListServer(command.Lister):
             default=False,
             help=_('List additional fields in output'),
         )
-        parser.add_argument(
+        name_lookup_group = parser.add_mutually_exclusive_group()
+        name_lookup_group.add_argument(
             '-n', '--no-name-lookup',
             action='store_true',
             default=False,
-            help=_('Skip flavor and image name lookup.'),
+            help=_('Skip flavor and image name lookup.'
+                   'Mutually exclusive with "--name-lookup-one-by-one"'
+                   ' option.'),
+        )
+        name_lookup_group.add_argument(
+            '--name-lookup-one-by-one',
+            action='store_true',
+            default=False,
+            help=_('When looking up flavor and image names, look them up'
+                   'one by one as needed instead of all together (default). '
+                   'Mutually exclusive with "--no-name-lookup|-n" option.'),
         )
         parser.add_argument(
             '--marker',
@@ -1224,28 +1277,46 @@ class ListServer(command.Lister):
                                            limit=parsed_args.limit)
 
         images = {}
-        # Create a dict that maps image_id to image object.
-        # Needed so that we can display the "Image Name" column.
-        # "Image Name" is not crucial, so we swallow any exceptions.
-        if data and not parsed_args.no_name_lookup:
-            try:
-                images_list = self.app.client_manager.image.images.list()
-                for i in images_list:
-                    images[i.id] = i
-            except Exception:
-                pass
-
         flavors = {}
-        # Create a dict that maps flavor_id to flavor object.
-        # Needed so that we can display the "Flavor Name" column.
-        # "Flavor Name" is not crucial, so we swallow any exceptions.
         if data and not parsed_args.no_name_lookup:
-            try:
-                flavors_list = compute_client.flavors.list(is_public=None)
-                for i in flavors_list:
-                    flavors[i.id] = i
-            except Exception:
-                pass
+            # Create a dict that maps image_id to image object.
+            # Needed so that we can display the "Image Name" column.
+            # "Image Name" is not crucial, so we swallow any exceptions.
+            # The 'image' attribute can be an empty string if the server was
+            # booted from a volume.
+            if parsed_args.name_lookup_one_by_one or image_id:
+                for i_id in set(filter(lambda x: x is not None,
+                                       (s.image.get('id') for s in data
+                                        if s.image))):
+                    try:
+                        images[i_id] = image_client.images.get(i_id)
+                    except Exception:
+                        pass
+            else:
+                try:
+                    images_list = image_client.images.list()
+                    for i in images_list:
+                        images[i.id] = i
+                except Exception:
+                    pass
+
+            # Create a dict that maps flavor_id to flavor object.
+            # Needed so that we can display the "Flavor Name" column.
+            # "Flavor Name" is not crucial, so we swallow any exceptions.
+            if parsed_args.name_lookup_one_by_one or flavor_id:
+                for f_id in set(filter(lambda x: x is not None,
+                                       (s.flavor.get('id') for s in data))):
+                    try:
+                        flavors[f_id] = compute_client.flavors.get(f_id)
+                    except Exception:
+                        pass
+            else:
+                try:
+                    flavors_list = compute_client.flavors.list(is_public=None)
+                    for i in flavors_list:
+                        flavors[i.id] = i
+                except Exception:
+                    pass
 
         # Populate image_name, image_id, flavor_name and flavor_id attributes
         # of server objects so that we can display those columns.
